from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable
from logging import warn
from math import inf, log
from typing import Callable, Iterable, List
from uuid import uuid4

from Akatosh.universe import Mundus

from .event import ContinuousEvent, Event, InstantEvent
from .logger import logger
from .resource import Resource
from .states import State


class Entity:
    def __init__(
        self,
        label: str | None = None,
        create_at: int
        | float
        | Callable[..., int]
        | Callable[..., float]
        | None = None,
        terminate_at: int
        | float
        | Callable[..., int]
        | Callable[..., float]
        | None = None,
        precursor: Entity | List[Entity] | None = None,
    ) -> None:
        """Create a entity.

        Args:
            label (str | None, optional): short description of the entity. Defaults to None.
            create_at (int | float | Callable | None, optional): when the life cycle of this entity should start. Defaults to 0, then must call create() method manually.
            terminate_at (int | float | Callable | None, optional): when the life cycle of this entity should end. Defaults to inf, then must call terminate() method manually.
        """
        self._id = uuid4().int
        self._label = label
        self._state: List[str] = list()
        self._occupied_resources: List[Resource] = list()
        self._registered_lists: List[EntityList] = list()
        self._creation: InstantEvent = None  # type: ignore
        self._termination: InstantEvent = None  # type: ignore
        self._events: List[Event] = list()

        # assign precursor and followers
        if isinstance(precursor, list):
            self._precursor = precursor
        elif precursor is None:
            self._precursor = []
        else:
            self._precursor = [precursor]
        self._followers: List[Entity] = []
        for event in self.precursor:
            event._followers.append(self)

        # assign create_at and terminate_at
        if create_at is not None:
            if callable(create_at):
                self._create_at = round(create_at(), Mundus.resolution)
            else:
                self._create_at = round(create_at, Mundus.resolution)
        else:
            self._create_at = 0
        self._created_at = float()

        if terminate_at is not None:
            if callable(terminate_at):
                self._terminate_at = round(terminate_at(), Mundus.resolution)
            else:
                self._terminate_at = round(terminate_at, Mundus.resolution)
        else:
            self._terminate_at = inf
        self._terminated_at = float()

        # create the entity if no precursor
        if create_at is not None and len(self.precursor) == 0:
            self.create(self.create_at)

    def create(self, at: int | float, force=False) -> None:
        """The creation of the entity."""

        # check if the entity is already over due
        if at > self.terminate_at:
            logger.debug(f"Entity {self.label} passed due time.")
            return

        # call back function for creation
        def _create():
            if self.terminated:
                raise RuntimeError(f"Entity {self.label} is already terminated.")
            if self.created:
                return
            self._created_at = Mundus.now
            self._state.append(State.CREATED)
            self.on_creation()
            logger.debug(f"Entity {self.label} created at {at}")

        # force creation if force is True, regardless of the precursor and the time
        if force:
            self._creation = InstantEvent(
                at=self.create_at if self.create_at > at else at,
                action=_create,
                label=f"Creation of {self.label}",
                priority=-2,
            )
            if self.terminate_at < inf:
                self.terminate(self.terminate_at)
        else:
            # check if all precursors are terminated, if so, create the entity
            if len(self.precursor) != 0:
                if all(p.terminated for p in self.precursor):
                    self._creation = InstantEvent(
                        at=self.create_at if self.create_at > at else at,
                        action=_create,
                        label=f"Creation of {self.label}",
                        priority=-2,
                    )
                    if self.terminate_at < inf:
                        self.terminate(self.terminate_at)
            # no precursor, create the entity
            else:
                self._creation = InstantEvent(
                    at=self.create_at if self.create_at > at else at,
                    action=_create,
                    label=f"Creation of {self.label}",
                    priority=-2,
                )
                if self.terminate_at < inf:
                    self.terminate(self.terminate_at)

    @abstractmethod
    def on_creation(self):
        """Callback function upon creation of the entity"""
        pass

    def terminate(self, at: int | float) -> None:
        """The termination of the entity. This will release all occupied resource, remove the entity from all entity lists, and cancel all unfinished events."""

        def _terminate():
            if self.terminated:
                return
            self._terminated_at = Mundus.now
            self._state.append(State.TERMINATED)
            self.release_resources()
            self.unregister_from_lists()
            self.cancel_unfinished_events()
            if self.created:
                self.on_termination()
            logger.debug(f"Entity {self.label} terminated at {Mundus.now}")
            for entity in self._followers:
                entity.create(Mundus.now)
                
        if self._termination is not None:
            self._termination.cancel()

        self._termination = InstantEvent(
            at=self.terminate_at if self.terminate_at < at else at,
            action=_terminate,
            label=f"Termination of {self.label}",
            priority=-2,
        )

    @abstractmethod
    def on_termination(self):
        """Callback function upon termination of the entity"""
        pass

    def get(
        self, resource: Resource, amount: int | float | Callable | None = None
    ) -> None:
        """Consume certain amout from a resource. if no amount is specified, consume all."""

        if not self.created:
            raise RuntimeError(
                f"Entity {self} is not created yet, can not interact with any resource."
            )

        if self.terminated:
            raise RuntimeError(
                f"Entity {self} is already terminated, can not interact with any resource."
            )

        if amount:
            if callable(amount):
                resource.distribute(self, amount())
                self.ocupied_resources.append(resource)
            else:
                resource.distribute(self, amount)
                self.ocupied_resources.append(resource)
        else:
            resource.distribute(self, resource.amount)

    def put(
        self, resource: Resource, amount: int | float | Callable | None = None
    ) -> None:
        """Return certain amout to a resource. if no amount is specified, return all."""
        if not self.created:
            raise RuntimeError(
                f"Entity {self} is not created yet, can not interact any resource."
            )

        if self.terminated:
            raise RuntimeError(
                f"Entity {self} is already terminated, can not interact any resource."
            )

        if amount:
            if callable(amount):
                resource.collect(self, amount())
                if self not in resource.users:
                    self.ocupied_resources.remove(resource)
            else:
                resource.put(amount)
                if self not in resource.users:
                    self.ocupied_resources.remove(resource)
        else:
            resource.put(resource.amount)

    def release_resources(self):
        """Release all occupied resources."""
        for res in self.ocupied_resources:
            res.collect(self)

    def unregister_from_lists(self):
        """Remove this entity from all registered entity lists."""
        for list in self.registered_lists[:]:
            list.remove(self)

    def cancel_unfinished_events(self):
        """Cancel all unfinished events."""
        for event in self.events:
            if not event.ended:
                event.cancel()

    def continuous_event(
        self,
        at: int | float | Callable,
        interval: int | float | Callable,
        duration: int | float | Callable,
        precursor: Event | List[Event] | None = None,
        priority: int | float | Callable = 0,
        label: str | None = None,
        **kwargs,
    ):
        """A decorator for creating continous event engaged by this entity.

        Args:
            at (int | float | Callable): when the continuous event starts.
            interval (int | float | Callable): how frequent the event happens.
            duration (int | float | Callable): the duration of the event.
            precursor (Event | List[Event] | None, optional): the precursors for this event. Defaults to None.
            priority (int | float | Callable, optional): the priority for this event. Defaults to 0.
            label (str | None, optional): short description of the event. Defaults to None.
        """

        def _continous_event(func: Callable):
            # define a function that validate the entity state first, then create the event
            def __continous_event():
                if not self.created:
                    raise RuntimeError(
                        f"Entity {self} is not created yet, can not engage in any event."
                    )
                if self.terminated:
                    raise RuntimeError(
                        f"Entity {self} is already terminated, can not engage in any event."
                    )
                self.events.append(
                    ContinuousEvent(
                        at=at,
                        interval=interval,
                        duration=duration,
                        precursor=precursor,
                        action=func,
                        priority=priority,
                        label=label,
                        **kwargs,
                    )
                )

            # create the instant event that will validate the entity state first, then create the continuous event
            InstantEvent(
                at=at,
                action=__continous_event,
                priority=-1,
                label=f"Engagement: {label}",
            )

        return _continous_event

    def instant_event(
        self,
        at: int | float | Callable,
        precursor: Event | List[Event] | None = None,
        priority: int | float | Callable = 0,
        label: str | None = None,
        **kwargs,
    ):
        """Decorator for creating instant event.

        Args:
            at (int | float | Callable): When the event happens.
            precursor (Event | List[Event] | None, optional): The precursor events. Defaults to None.
            priority (int | float | Callable, optional): The priority of the event. Defaults to 0.
            label (str | None, optional): Short description of the event. Defaults to None.
        """

        def _instant_event(func: Callable):
            def __instant_event():
                if not self.created:
                    raise RuntimeError(
                        f"Entity {self} is not created yet, can not engage in any event."
                    )
                if self.terminated:
                    raise RuntimeError(
                        f"Entity {self} is already terminated, can not engage in any event."
                    )

                self.events.append(
                    InstantEvent(
                        at=at,
                        precursor=precursor,
                        action=func,
                        priority=priority,
                        label=label,
                        **kwargs,
                    )
                )

            # create the instant event that will validate the entity state first, then create the event
            InstantEvent(
                at=at, action=__instant_event, priority=-1, label=f"Engagement: {label}"
            )

        return _instant_event

    @property
    def label(self):
        """Return the label of the entity."""
        return self._label

    @property
    def state(self):
        """Return the state(s) of the entity."""
        return self._state

    @property
    def created(self):
        """Return True if the entity is created."""
        return State.CREATED in self.state

    @property
    def create_at(self):
        """Return the time when the entity should be created."""
        return self._create_at
    
    @property
    def created_at(self):
        """Return the time when the entity is created."""
        return self._created_at

    @property
    def terminated(self):
        """Return True if the entity is terminated."""
        return State.TERMINATED in self.state

    @property
    def terminate_at(self):
        """Return the time when the entity should be terminated."""
        return self._terminate_at

    @property
    def terminated_at(self):
        """Return the time when the entity is terminated."""
        return self._terminated_at
    
    @property
    def ocupied_resources(self):
        """Return the resources occupied by the entity."""
        return self._occupied_resources

    @property
    def registered_lists(self):
        """Return the entity lists contains the entity."""
        return self._registered_lists

    @property
    def events(self):
        """Return the events engaged by the entity."""
        return self._events

    @property
    def precursor(self):
        """Return the precursor(s) of the entity."""
        return self._precursor


class EntityList(list):
    """Customized list for entities. This list ensures all items are entities and unique. When an entity is terminated, it will be removed from all entity lists automatically."""

    def __init__(self, iterable: Iterable[Entity] = [], label: str | None = None):
        """Create a list for entities, with optional label.

        Args:
            iterable (Iterable[Entity]): Iterable of entities.
        """
        super().__init__(item for item in iterable if isinstance(item, Entity))
        self._label = label

    def insert(self, __index: int, __object: Entity) -> None:
        """Insert an entity to the list.

        Args:
            __index (SupportsIndex): the index to insert the entity.
            __object (Entity): the entity to be inserted.
        """
        if __object not in self:
            super().insert(__index, __object)
            if self not in __object.registered_lists:
                __object.registered_lists.append(self)
            logger.debug(
                f"Entity {__object.label} is inserted to {self.label if self.label else self} at {__index}."
            )

    def append(self, __object: Entity) -> None:
        """Append an entity to the list.

        Args:
            __object (Entity): the entity to be appended.
        """
        if __object not in self:
            super().append(__object)
            if self not in __object.registered_lists:
                __object.registered_lists.append(self)
            logger.debug(
                f"Entity {__object.label} is appended to {self.label if self.label else self}."
            )

    def remove(self, __object: Entity) -> None:
        """Remove an entity from the list.

        Args:
            __object (Entity): the entity to be removed.
        """
        super().remove(__object)
        __object.registered_lists.remove(self)
        logger.debug(
            f"Entity {__object.label} is removed from {self.label if self.label else self}."
        )

    def pop(self, __index: int = ...) -> Entity:
        """Pop an entity from the list.

        Args:
            __index (int, optional): the index. Defaults to ... (the last one).

        Returns:
            Entity: return the poped entity.
        """
        __object: Entity = super().pop(__index)
        __object.registered_lists.remove(self)
        logger.debug(
            f"Entity {__object.label} is poped from {self.label if self.label else self} at {__index}."
        )
        return __object

    def clear(self) -> None:
        """Clear the list."""
        for item in self:
            item.registered_lists.remove(self)
        super().clear()
        logger.debug(f"{self.label if self.label else self} is cleared.")

    def extend(self, __iterable: Iterable[Entity]) -> None:
        """Extend the list with an iterable of entities.

        Args:
            __iterable (Iterable[Entity]): the iterable of entities.
        """
        for item in __iterable:
            if item not in self:
                super().append(item)
        for item in __iterable:
            if self not in item.registered_lists:
                item.registered_lists.append(self)
        logger.debug(
            f"{self.label if self.label else self} is extended with {__iterable}."
        )

    @property
    def label(self):
        """Return the label of the entity list."""
        return self._label



from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import Any, Callable, List
from uuid import uuid4
import warnings

from .logger import logger
from .states import State
from .universe import Mundus


class Event:
    def __init__(
        self,
        at: int | float | Callable[..., int] | Callable[..., float],
        precursor: Event | List[Event] | None = None,
        action: Callable | None = None,
        priority: int | float | Callable[..., int] | Callable[..., float] = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Base class for all events.

        Args:
            at (int | float | Callable[...,int] | Callable[...,float]): the time when the event happens.
            precursor (Event | List[Event] | None, optional): the precursor events. Defaults to None.
            action (Callable | None, optional): the actual action of the event, must be defined. Defaults to None.
            priority (int | float | Callable[...,int] | Callable[...,float], optional): the priority of the event. When multiple event happen at the same time, the event with lower priority value will happen first. Defaults to 0.
            label (str | None, optional): Short descirption of the event. Defaults to None.
        """
        self._id = uuid4().int
        # set the time
        if callable(at):
            self._at = round(at(), Mundus.resolution)
        else:
            self._at = round(at, Mundus.resolution)
        # set the target event(s) that this event should wait and follow
        if isinstance(precursor, list):
            self._precursor = precursor
        elif precursor is None:
            self._precursor = []
        else:
            self._precursor = [precursor]
        self._follower: List[Event] = []
        for event in self.precursor:
            event._follower.append(self)
        # set the state
        if len(self.precursor) != 0:
            self.state = State.INACTIVE
        else:
            self.state = State.ACTIVE
        # set the action and the priority
        if action:
            self._action = action
        if callable(priority):
            self._priority = priority()
        else:
            self._priority = priority
        # assign label
        self._label = label
        # add to the universe
        if Mundus.now < self.at:
            Mundus.future_events.append(self)
        elif Mundus.now > self.at:
            raise RuntimeError(f"Event {self.label} tries to alter the past.")
        else:
            Mundus.current_events.append(self)

    def end(self):
        """End the event and activate the follower events if there is any."""
        if self.ended:
            return

        self.state = State.ENDED
        if self in Mundus.future_events:
            Mundus.future_events.remove(self)
        if self in Mundus.current_events:
            Mundus.current_events.remove(self)
        Mundus.past_events.append(self)
        for event in self.follower:
            try:
                event.activate()
            except RuntimeError:
                logger.debug(f"Event {event.label} passed due time.")
        logger.debug(f"Event {self.label} is ended.")

    def activate(self, force: bool = False):
        """Activate the event."""

        # raise error if the event has already ended
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")

        # raise warning if the event is cancelled
        if self.cancelled:
            warnings.warn(f"Event {self.label} is cancelled.")
            return

        # return if the event is already active
        if self.active:
            return

        if force:
            self.state = State.ACTIVE
            if self._at < Mundus.now:
                self._at = Mundus.now
            logger.debug(f"Event {self.label} is set to active.")
        else:
            if all(e.state == State.ENDED for e in self.precursor):
                self.state = State.ACTIVE
                if self._at < Mundus.now:
                    self._at = Mundus.now
                logger.debug(f"Event {self.label} is set to active.")
            else:
                warnings.warn(f"Event {self.label} is waiting for other events to end.")
                return

    def deactivate(self):
        """Deactivate the event.

        Raises:
            RuntimeError: raise if the event has already ended.
        """

        # raise error if the event has already ended
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")

        # raise warning if the event is cancelled
        if self.cancelled:
            warnings.warn(f"Event {self.label} is cancelled.")
            return

        # return if the event is already inactive
        if self.inactive:
            return

        self.state = State.INACTIVE

    def cancel(self):
        """Cancel the event. Will not set the follower events to active.

        Raises:
            RuntimeError: raise if the event has already ended.
        """

        # raise error if the event has already ended
        if self.ended:
            raise RuntimeError(f"Event {self.label} has already ended.")

        # return if the event is already cancelled
        if self.cancelled:
            return

        self.state = State.CANCELED
        if self in Mundus.future_events:
            Mundus.future_events.remove(self)
        if self in Mundus.current_events:
            Mundus.current_events.remove(self)
        Mundus.past_events.append(self)
        logger.debug(f"Event {self.label} is cancelled.")

    @abstractmethod
    async def _perform(self):
        """Abstract method for the event to perform its action."""
        pass

    def __eq__(self, _o: Event) -> bool:
        """Determine if two events are the same."""
        return self.id == _o.id

    @property
    def id(self) -> int:
        """Return the unique id of the event."""
        return self._id

    @property
    def label(self) -> str | None:
        """Return the label of the event."""
        return self._label

    @property
    def at(self) -> int | float:
        """Return the time when the event happens."""
        return self._at

    @property
    def precursor(self) -> List[Event]:
        """Return the precursor events."""
        return self._precursor

    @property
    def follower(self) -> List[Event]:
        """Return the follower events."""
        return self._follower

    @property
    def action(self) -> Callable | None:
        """Return the action of the event."""
        return self._action

    @property
    def priority(self) -> int | float:
        """Return the priority of the event."""
        return self._priority

    @property
    def state(self) -> str:
        """Return the state of the event."""
        return self._state

    @state.setter
    def state(self, state: str):
        """Set the state of the event."""
        self._state = state

    @property
    def active(self) -> bool:
        """Return True if the event is active."""
        return State.ACTIVE == self.state

    @property
    def inactive(self) -> bool:
        """Return True if the event is inactive."""
        return State.INACTIVE == self.state

    @property
    def cancelled(self) -> bool:
        """Return True if the event is cancelled."""
        return State.CANCELED == self.state

    @property
    def ended(self) -> bool:
        """Return True if the event is ended."""
        return State.ENDED == self.state


class InstantEvent(Event):
    """Instant event is an event that happens at a specific time and only happens once."""

    def __init__(
        self,
        at: int | float | Callable[..., int] | Callable[..., float],
        precursor: Event | List[Event] | None = None,
        action: Callable[..., Any] | None = None,
        priority: int | float | Callable[..., int] | Callable[..., float] = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Create a instant event.

        Args:
            at (int | float | Callable[...,int] | Callable[...,float]): the time when the event happens.
            precursor (Event | List[Event] | None, optional): the precursor events. Defaults to None.
            action (Callable | None, optional): the actual action of the event, must be defined. Defaults to None.
            priority (int | float | Callable[...,int] | Callable[...,float], optional): the priority of the event. When multiple event happen at the same time, the event with lower priority value will happen first. Defaults to 0.
            label (str | None, optional): Short descirption of the event. Defaults to None.
        """
        super().__init__(at, precursor, action, priority, label, **kwargs)

    async def _perform(self):
        """Perform the action of the event."""
        if self.state == State.ACTIVE:
            if self.action:
                if inspect.iscoroutinefunction(self.action):
                    await self.action()
                else:
                    self.action()
            Mundus.current_events.remove(self)
            Mundus.past_events.append(self)
            self.end()


def instant_event(
    at: int | float | Callable[..., int] | Callable[..., float],
    precursor: Event | List[Event] | None = None,
    priority: int | float | Callable[..., int] | Callable[..., float] = 0,
    label: str | None = None,
    **kwargs,
):
    """Decorator for creating instant event.

    Args:
        at (int | float | Callable): When the event happens.
        precursor (Event | List[Event] | None, optional): The precursor events. Defaults to None.
        priority (int | float | Callable, optional): The priority of the event. Defaults to 0.
        label (str | None, optional): Short description of the event. Defaults to None.
    """

    def _instant_event(func: Callable):
        return InstantEvent(
            at=at,
            precursor=precursor,
            action=func,
            priority=priority,
            label=label,
            **kwargs,
        )

    return _instant_event


class ContinuousEvent(Event):
    def __init__(
        self,
        at: int | float | Callable[..., int] | Callable[..., float],
        interval: int | float | Callable[..., int] | Callable[..., float],
        duration: int | float | Callable[..., int] | Callable[..., float],
        precursor: Event | List[Event] | None = None,
        action: Callable[..., Any] | None = None,
        priority: int | float | Callable[..., int] | Callable[..., float] = 0,
        label: str | None = None,
        **kwargs,
    ) -> None:
        """Continous event is an event that happens at a specific time and happens multiple times for a specific duration.

        Args:
            at (int | float | Callable[...,int] | Callable[...,float]): when the event happens.
            interval (int | float | Callable[...,int] | Callable[...,float]): the interval between each action repeat.
            duration (int | float | Callable[...,int] | Callable[...,float]): the duration of the event.
            precursor (Event | List[Event] | None, optional): the precursor events, can be both instant events or continous events. Defaults to None.
            action (Callable[..., Any] | None, optional): the actual action of the event. Defaults to None.
            priority (int | float | Callable[...,int] | Callable[...,float], optional): the priority of the event. Defaults to 0.
            label (str | None, optional): short description of the event.. Defaults to None.
        """
        super().__init__(
            at=at,
            precursor=precursor,
            action=action,
            priority=priority,
            label=label,
            **kwargs,
        )

        self._interval = interval
        if callable(duration):
            self._duration = round(duration(), Mundus.resolution)
        else:
            self._duration = round(duration, Mundus.resolution)
        self._till = self.at + self.duration
        self._sub_events: List[InstantEvent] = list()

    async def _perform(self):
        """Perform the action of the event. Compare to instant event, continous event will perform the action multiple times."""
        if self.action:
            self.action()
            if callable(self.interval):
                self._at = round(self.interval() + Mundus.now, Mundus.resolution)
            else:
                self._at = round(self.interval + Mundus.now, Mundus.resolution)
            if self.at <= self.till:
                logger.debug(f"Event {self.label} next step is at {self.at}.")
                Mundus.future_events.append(self)
                Mundus.current_events.remove(self)
            else:
                self.end()

    def activate(self, force: bool = False):
        if Mundus.now > self.till:
            logger.debug(f"Event {self.label} has passed due time.")
            return
        else:
            super().activate(force)

    @property
    def interval(self) -> int | float | Callable[..., Any]:
        """Return the interval between each action repeat."""
        return self._interval

    @property
    def duration(self) -> int | float:
        """Return the duration of the event."""
        return self._duration

    @property
    def till(self) -> int | float:
        """Return the time when the event ends."""
        return self._till

    @property
    def sub_events(self) -> List[InstantEvent]:
        """Return the sub events of the continous event."""
        return self._sub_events


def continuous_event(
    at: int | float | Callable[..., int] | Callable[..., float],
    interval: int | float | Callable[..., int] | Callable[..., float],
    duration: int | float | Callable[..., int] | Callable[..., float],
    precursor: Event | List[Event] | None = None,
    priority: int | float | Callable[..., int] | Callable[..., float] = 0,
    label: str | None = None,
    **kwargs,
):
    """A decorator for creating continous event.

    Args:
        at (int | float | Callable[...,int] | Callable[...,float]): when the continuous event starts.
        interval (int | float | Callable[...,int] | Callable[...,float]): how frequent the event happens.
        duration (int | float | Callable[...,int] | Callable[...,float]): the duration of the event.
        precursor (Event | List[Event] | None, optional): the precursors for this event. Defaults to None.
        priority (int | float | Callable[...,int] | Callable[...,float], optional): the priority for this event. Defaults to 0.
        label (str | None, optional): short description of the event. Defaults to None.
    """

    def _continous_event(func: Callable):
        return ContinuousEvent(
            at=at,
            interval=interval,
            duration=duration,
            precursor=precursor,
            action=func,
            priority=priority,
            label=label,
            **kwargs,
        )

    return _continous_event



from .entity import Entity, EntityList
from .event import (ContinuousEvent, Event, InstantEvent, continuous_event,
                    instant_event)
from .resource import Resource
from .states import State
from .universe import Mundus



import logging

logger = logging.getLogger("Akatosh")
formatter = logging.Formatter("%(asctime)s\t%(levelname)s\t%(message)s")

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

file_handler = logging.FileHandler(filename=".log", mode="w")
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)


from __future__ import annotations

import asyncio
import logging
from math import log
from typing import TYPE_CHECKING, List

from .logger import logger
from .states import State

if TYPE_CHECKING:
    from .event import Event


class Universe:
    def __init__(self) -> None:
        """The Simulation Universe."""
        self._resolution = 15  # the resolution of the time. 1 means minimum time unit is 0.1 second.
        self._now: int | float = -1  # the current time of the simulated universe.
        self._future_events: List[Event] = list()  # the future events queue.
        self._current_events: List[Event] = list()  # the current events queue.
        self._past_events: List[Event] = list()  # the past events queue.
        self._alduin: bool = False  # trigger to stop the simulation.
        self.set_logger(logging.ERROR)  # set the default logger level to ERROR.

    async def akatosh(self, till: int | float | None = None):
        """Akatosh is the god of time in the Elder Scrolls universe. This method is the core of the simulation.

        Args:
            till (int | float | None, optional): the end time of the simulated universe. Defaults to None.

        Raises:
            RuntimeError: raise if the event is in unknown state or being placed in wrong event queue.
        """
        while len(self.future_events) != 0:
            
            # stop simulation if world eater alduin is triggered.
            if self.alduin:
                return

            # update the current time of the simulated universe.
            valid_future_events = [
                event
                for event in self.future_events
                if event.state == State.ACTIVE or event.state == State.INACTIVE
            ]
            if len(valid_future_events) == 0:
                return
            _time = min(event.at for event in valid_future_events)
            if self.now < _time:
                self._now = _time
                
            # stop simulation if the end time is reached.
            if till is not None:
                if self.now > till:
                    return
                
            logger.debug(f"Time: {self.now}")

            logger.debug(
                f"Future active events: {[(event.at, event.label) for event in self.future_events]}"
            )

            self.organise_events()

            logger.debug(
                f"Current events: {[(event.priority, event.label) for event in self.current_events]}"
            )

            await self.execute_current_events()
            
            

    def organise_events(self):
        """Organise the events in the future events queue. Valid event will be moved to the current events queue. Non-valid event will be moved to the past events queue.

        Raises:
            RuntimeError: raise if the event is in unknown state or being placed in wrong event queue.
        """
        for event in self.future_events[:]:
            if event.at <= self.now:
                if event.state == State.ACTIVE or event.state == State.INACTIVE:
                    logger.debug(f"Event {event.label} is triggered.")
                    self.future_events.remove(event)
                    self.current_events.append(event)
                elif event.state == State.CANCELED:
                    raise RuntimeError(
                            f"Event {event.label} is canceled but inside future events queue."
                        )
                elif event.state == State.ENDED:
                    raise RuntimeError(
                            f"Event {event.label} is ended but inside future events queue."
                        )
                else:
                    raise RuntimeError(f"Event {event.label} is in unknown state.")

    async def execute_current_events(self):
        """Execute the current events queue based on priority. The event with the lowest priority will be executed first. The event with the same priority will be executed concurrently.
        """
        while (
            len([event for event in self.current_events if event.state == State.ACTIVE])
            != 0
        ):
            priority = min(
                event.priority
                for event in self.current_events
                if event.state == State.ACTIVE
            )
            logger.debug(
                f"Current Priority: {priority}, Executing events: {[event.label for event in self.current_events if event.priority == priority]}"
            )
            await asyncio.gather(
                *[
                    event._perform()
                    for event in self.current_events
                    if event.priority == priority
                ]
            )

    def simulate(self, till: int | float | None = None):
        """Start the simulation."""
        asyncio.run(self.akatosh(till))

    def set_logger(self, level: int):
        """Set the logger level."""
        logger.setLevel(level)

    @property
    def resolution(self) -> int:
        """Return  the resolution of the time."""
        return self._resolution
    
    @resolution.setter
    def resolution(self, value: int):
        """Set the resolution of the time."""
        if value <= 0:
            raise ValueError("Resolution must be greater than 0.")
        self._resolution = value

    @property
    def now(self) -> int | float:
        """Return the current time of the simulated universe."""
        return self._now

    @property
    def future_events(self) -> List[Event]:
        """Return the future events queue."""
        return self._future_events

    @property
    def current_events(self) -> List[Event]:
        """Return the current events queue."""
        return self._current_events

    @property
    def past_events(self) -> List[Event]:
        """Return the past events queue."""
        return self._past_events

    @property
    def alduin(self) -> bool:
        """Return the trigger to stop the simulation."""
        return self._alduin

    @alduin.setter
    def alduin(self, value: bool):
        """Set the trigger to stop the simulation."""
        self._alduin = value


Mundus = Universe()
